import os
import importlib
from typing import List
from colorama import Fore
from gupi.errors import GupiBaseException
from gupi.custom_logger import getLogger
from gupi.error_handler import error_handler

from tools.analyser import _analyse

logger = getLogger("gupi")


@error_handler
def run(vendors: List[str]):
    pkg_info, orders = _analyse(vendors)
    for vendor in vendors:
        for pkg_name in orders[vendor]:
            pkg = pkg_info[vendor][pkg_name]
            main_mod = f'{vendor}.{pkg_name}.main'
            main_path = f'{vendor}/{pkg_name}/main.py'
            try:
                assert os.path.isfile(main_path)
                module_obj = importlib.import_module(main_mod)
                logger.info(f'Loaded module \'{pkg_name}\'')
            except AssertionError:
                pass
            except ImportError as e:
                raise GupiBaseException(
                    title=f'From module {Fore.CYAN}{pkg["name"]}{Fore.RESET}:',
                    message=e.msg + '.',
                    exc_info=(type(e), e, e.__traceback__)
                )




import os
import traceback
import functools

import gupi.common as common
from gupi.errors import GupiBaseException
from gupi.custom_logger import getLogger, log


def error_handler(fun=None, *, reraise=False):
    if fun is None:
        return lambda f: error_handler(f, reraise=reraise)

    @functools.wraps(fun)
    def wrapper(*args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except GupiBaseException as e:
            tb = traceback.extract_tb(e.__traceback__)
            filename, lineno, func, text = tb[-1]

            filename = os.path.relpath(filename, common.REPO_PATH)
            filename = filename.split('.')[0].replace('/', '.')
            
            logger = getLogger(filename)
            log(logger.error, e.message, title=e.title, exc_info=e.exc_info)
            
            if reraise: raise

    return wrapper



from colorama import Fore
from typing import Tuple, List, Optional

from gupi.common import hili


class GupiBaseException(Exception):
    def __init__(
        self,
        message: str,
        title: Optional[str] = None,
        exc_info = None,
        hint: Optional[str] = None
    ):
        self.message = message
        self.hint = hint
        self.title = title
        self.exc_info = exc_info
        super().__init__(message)

class DepVersionException(GupiBaseException):
    def __init__(self, org_pkg: dict, dep_pkg: dict, dependency: Tuple[str]):
        super().__init__(
            title=f'From package {Fore.CYAN}{org_pkg["name"]}{Fore.RESET}:',
            message=f"Package {hili(dep_pkg['name'])} "
            f"(v{dep_pkg.get('version')}) doesn't match with "
            f'required ({dependency[1]}{dependency[2]}).'
        )

class DepMissingException(GupiBaseException):
    def __init__(self, pkg: dict, dependency: Tuple[str], reason: str):
        self.reason = reason
        super().__init__(
            title=f'Package {hili(pkg["name"])}',
            message=f'depends on a {reason} package {hili(dependency[0])}.'
        )

class DepCycleException(GupiBaseException):
    def __init__(self, pkg_name: str, traceback: List[str]):
        self.pkg_name = pkg_name
        self.traceback = traceback
        super().__init__(
            title=f'Package {hili(pkg_name)}',
            message=(
                'creates a dependency cycle while trying to load it.\n'
                f'Traceback: {" -> ".join(traceback)}'
            )
        )

class VendorNotDirectory(GupiBaseException):
    def __init__(self, vendor: str):
        super().__init__(
            f'{hili(vendor)} is not a directory!',
            hint=(
                f'This issue usually occurs if {hili(vendor)}'
                'neither exists or presents as a folder in '
                'this repository path.'
            )
        )

class InvalidInfoJsonSyntax(GupiBaseException):
    def __init__(self, pkg_name: str):
        super().__init__(
            title=f'Package {hili(pkg_name)}',
            message='has a syntax-invalid info.json file'
        )

class InvalidInfoField(GupiBaseException):
    def __init__(self, pkg_name: str, invalid_field: str):
        self.field = invalid_field
        super().__init__(
            title=f'Package {hili(pkg_name)}',
            message=(
                f"has an invalid field '{self.field}'"
                ' in its info.json file.'
            ),
            hint=(
                'Remove this field manually or check if '
                "there's any typo in it."
            )
        )

class MissingInfoField(GupiBaseException):
    def __init__(self, pkg_name: str, missing_field: str):
        self.field = missing_field
        super().__init__(
            title=f'Package {hili(pkg_name)}',
            message=f"is missing a required field: '{self.field}'"
        )



import os
import re
import json

from colorama import Fore, Style
from typing import List, Tuple, Optional

import gupi.common as common
from gupi.errors import *
from gupi.error_handler import error_handler
from gupi.custom_logger import getLogger, log

logger = getLogger("gupi")

REQUIRED_FIELDS = {'name', 'version'}
ACCEPTED_FIELDS = {'name', 'dependencies', 'disabled', 'description', 'version'}


disableds = set()
pkg_info = {}
_processed = set()
_in_process = []


def parse_dependency(s):
    pattern = r"([a-zA-Z0-9_.-]+)([<>=!~]+)?([0-9.*]+)?"
    match = re.match(pattern, s)
    
    if match:
        name = match.group(1)
        type_ = match.group(2) if match.group(2) else None
        version = match.group(3) if match.group(3) else None
        return [name, type_, version]
    return [s, None, None]


def examine_dependency(
    vendor: str,
    org_pkg: dict,
    dependency: Tuple[str, Optional[str], Optional[str]]
):
    dep_pkg: dict = pkg_info[vendor][dependency[0]]

    if dependency[1] == '==' and dep_pkg.get('version') != dependency[2] \
    or dependency[1] == '>=' and dep_pkg.get('version') <  dependency[2] \
    or dependency[1] == '<=' and dep_pkg.get('version') >  dependency[2] \
    or dependency[1] == '>'  and dep_pkg.get('version') <= dependency[2] \
    or dependency[1] == '<'  and dep_pkg.get('version') >= dependency[2]:
        raise DepVersionException(org_pkg, dep_pkg, dependency)

    return dep_pkg


def topo_sort(vendor: str, pkg: dict, result: List[str]):
    _in_process.append(pkg['name'])

    for dependency in pkg['dependencies']:
        try:
            _depend = examine_dependency(vendor, pkg, dependency)
        except KeyError:
            _is_disabled = dependency[0] in disableds
            reason = "disabled" if _is_disabled else "non-existing"
            raise DepMissingException(pkg, dependency, reason)

        if _depend['name'] in _in_process:
            raise DepCycleException(_depend['name'], [p for p in _in_process] + [_depend['name']])

        if _depend['name'] not in _processed:
            topo_sort(vendor, _depend, result)
    
    _processed.add(pkg['name'])
    assert _in_process.pop() == pkg['name']

    result.append(pkg['name'])
    logger.info('Analysed %s', pkg['name'])


def _analyse(vendors: List[str]):
    """
    Returns:
        `(pkg_info, orders)`

        `pkg_info` and `orders` are both `dict[vendor_name, ...]`
    """
    orders = {}

    for vendor in vendors:
        logger.info(f'{f" {vendor.upper()} PACKAGES ":=^25}')

        pkg_info[vendor] = {}
        vendor_path = os.path.join(common.REPO_PATH, vendor)
        if not os.path.isdir(vendor_path):
            raise VendorNotDirectory(vendor_path)

        for pkg_name in os.listdir(vendor_path):
            pkg_path = os.path.join(vendor_path, pkg_name)
            if os.path.isfile(pkg_path): continue
            if pkg_name == '__pycache__': continue

            if '.' in pkg_name:
                log(logger.warning,
                    title=f'Package {Fore.CYAN}{pkg_name}{Fore.RESET}',
                    message=(
                        'contain a dot (.) character in its name '
                        f'({Fore.RED}skipped{Fore.RESET})'
                    )
                )
                continue

            pkg_info_path = os.path.join(pkg_path, 'info.json')

            try:
                with open(pkg_info_path, 'r', encoding='utf-8') as file:
                    pkg = json.load(file)
                    for field in REQUIRED_FIELDS:
                        if field not in pkg.keys():
                            raise MissingInfoField(pkg_name, field)
                    invalid_fields = [
                        field for field in pkg.keys()
                        if field not in ACCEPTED_FIELDS
                        and not field.startswith('_')
                    ]
                    if invalid_fields:
                        raise InvalidInfoField(pkg_name, invalid_fields[0])

                    assert pkg['name'] == pkg_name
                    if pkg.get('disabled', False):
                        disableds.add(pkg_name)
                        continue
                    pkg_info[vendor][pkg_name] = pkg
                
                pkg['dependencies'] = [
                    parse_dependency(dep)
                    for dep in pkg.get('dependencies', [])
                ]

            except FileNotFoundError as e:
                assert e.filename == pkg_info_path
                log(logger.warning,
                    title=f'Module {Fore.CYAN}{pkg_name}{Fore.RESET}',
                    message=(
                        'doesn\'t contain an info.json file '
                        f'({Fore.RED}skipped{Fore.RESET})'
                    )
                )
                continue

            except json.decoder.JSONDecodeError:
                raise InvalidInfoJsonSyntax(pkg_name)


        if pkg_info[vendor] == {}:
            logger.info(f'{Style.DIM}Nothing to load.{Style.RESET_ALL}')
            continue
        
        order = []
        for pkg in pkg_info[vendor].values():
            if pkg['name'] not in _processed:
                topo_sort(vendor, pkg, order)

        _processed.clear()
        _in_process.clear()

        orders[vendor] = order

    return pkg_info, orders


@error_handler
def analyse(vendors: List[str]):
    """
    Same as `_analyse`, but with the `error_handler` decorator.

    Returns:
        `(pkg_info, orders)`

        `pkg_info` and `orders` are both `dict[vendor_name, ...]`
    """
    return _analyse(vendors)



import logging

from functools import lru_cache
from colorama import Style, Fore, Back


_lf = f'{Back.LIGHTBLACK_EX}'
_rg = f'{Back.RESET}{Fore.RESET}'
lvlname = {
    'INFO':    f'     {_lf} INFO {_rg}',
    'DEBUG':   f'    {_lf} DEBUG {_rg}',
    'WARNING': f'  {_lf} WARNING {_rg}',
    'ERROR':   f'    {_lf} ERROR {_rg}',
}

class CustomFormatter(logging.Formatter):
    def format(self, record):
        log_format = (
            f'{lvlname[record.levelname]}  '
            f'{Fore.CYAN}'
            '{name} \033[1;37m| '
            f'{Style.RESET_ALL}'
        )

        if hasattr(record, 'title'):
            log_format += '{title}\n' + ' ' * (13 + len(record.name)) + ' | '
        log_format += '{message}'
        
        formatter = logging.Formatter(log_format, style='{')
        return formatter.format(record)

def log(f, message: str, *, title: str = None, exc_info=None):
    '''
    A custom log function tailored to use with title.
    >>> log(logger.info, 'Title', 'Message')
    '''
    if title is None:
        f(message, exc_info=exc_info)
    else:
        f(message, exc_info=exc_info, extra={'title': title})

@lru_cache(maxsize=1)
def _getHandler():
    handler = logging.StreamHandler()
    handler.setFormatter(CustomFormatter())
    return handler

_verbose_lvl = logging.INFO
def setVerbosity(_debug: bool):
    global _verbose_lvl
    if _debug:
        _verbose_lvl = logging.DEBUG

def getLogger(name: str):
    logger = logging.getLogger(name)
    logger.setLevel(_verbose_lvl)
    logger.addHandler(_getHandler())
    return logger



import argparse
from pathlib import Path

import gupi.common as common
from gupi.custom_logger import setVerbosity

parser = argparse.ArgumentParser()

def _load_args():
    subparsers = parser.add_subparsers(dest="command", required=True)
    parser.add_argument("repo_path", type=str, default=".", help=\
        "Path to repository (default to current)."
    )
    parser.add_argument("-v", "--verbose", action="store_true")

    analyse_prsr = subparsers.add_parser("analyse", help=\
        "Safe checking the current repository via info.json"
    )
    # fix_prsr = subparsers.add_parser("fix", help="Fix faults found by analyser's result")
    run_prsr = subparsers.add_parser("run", help="Run some module")

    args = parser.parse_args()
    return args

def main():
    args = _load_args()
    setVerbosity(args.verbose)
    common.REPO_PATH = str(Path(args.repo_path).resolve(True))

    try:
        with open("vendor.list", "r") as file:
            vendors = [
                vendor
                for vendor in file.read().splitlines()
                if vendor
            ]

    except FileNotFoundError:
        msg = (
            "File 'vendor.list' does not exist.\n"
            "Do you want to create one? (Y/n): "
        )
        if input(msg).strip().lower() in ('', 'y'):
            with open("vendor.list", "w") as file:
                print("Created.")
        exit()

    if args.command == "analyse":
        from tools import analyser
        analyser.analyse(vendors)
    elif args.command == "run":
        from tools import runner
        runner.run(vendors)

if __name__ == "__main__":
    main()


from setuptools import setup, find_packages

setup(
    name="gupi",
    version="1.0.6",
    packages=find_packages(),
    install_requires=[
        "colorama"
    ],
    entry_points={
        "console_scripts": [
            "gupi = gupi.cli:main",  # Lệnh `mycli` sẽ gọi hàm `main` trong `cli.py`
        ],
    },
    author="Dao Le Bao Minh",
    author_email="djnner@proton.me",
    description="Dependency analyser for my future Python projects",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/GlowCheese/gupi",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)
